package sun.nio.ch;

import java.nio.BufferOverflowException;
import sun.misc.SharedSecrets;
import java.nio.ByteBuffer;
import java.nio.channels.NonReadableChannelException;
import java.util.concurrent.Future;
import java.nio.channels.FileLock;
import java.nio.channels.CompletionHandler;
import java.nio.channels.NonWritableChannelException;
import java.nio.channels.AsynchronousCloseException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.AsynchronousFileChannel;
import java.io.IOException;
import java.io.FileDescriptor;
import sun.misc.JavaIOFileDescriptorAccess;

public class WindowsAsynchronousFileChannelImpl extends AsynchronousFileChannelImpl implements Iocp.OverlappedChannel, Groupable
{
    private static final JavaIOFileDescriptorAccess fdAccess;
    private static final int ERROR_HANDLE_EOF = 38;
    private static final FileDispatcher nd;
    private final long handle;
    private final int completionKey;
    private final Iocp iocp;
    private final boolean isDefaultIocp;
    private final PendingIoCache ioCache;
    static final int NO_LOCK = -1;
    static final int LOCKED = 0;
    
    private WindowsAsynchronousFileChannelImpl(FileDescriptor fileDescriptor, boolean b, boolean b2, Iocp iocp, boolean isDefaultIocp) throws IOException {
        super(fileDescriptor, b, b2, iocp.executor());
        this.handle = WindowsAsynchronousFileChannelImpl.fdAccess.getHandle(fileDescriptor);
        this.iocp = iocp;
        this.isDefaultIocp = isDefaultIocp;
        this.ioCache = new PendingIoCache();
        this.completionKey = iocp.associate(this, this.handle);
    }
    
    public static AsynchronousFileChannel open(FileDescriptor fileDescriptor, boolean b, boolean b2, ThreadPool threadPool) throws IOException {
        Iocp iocp;
        int n;
        if (threadPool == null) {
            iocp = WindowsAsynchronousFileChannelImpl.DefaultIocpHolder.defaultIocp;
            n = true;
        }
        else {
            iocp = new Iocp(null, threadPool).start();
            n = false;
        }
        try {
            return new WindowsAsynchronousFileChannelImpl(fileDescriptor, b, b2, iocp, (boolean)n);
        }
        catch (IOException ex) {
            if (n == 0) {
                iocp.implClose();
            }
            throw ex;
        }
    }
    
    @Override
    public <V, A> PendingFuture<V, A> getByOverlapped(long n) {
        return this.ioCache.remove(n);
    }
    
    @Override
    public void close() throws IOException {
        this.closeLock.writeLock().lock();
        try {
            if (this.closed) {
                return;
            }
            this.closed = true;
        }
        finally {
            this.closeLock.writeLock().unlock();
        }
        this.invalidateAllLocks();
        close0(this.handle);
        this.ioCache.close();
        this.iocp.disassociate(this.completionKey);
        if (!this.isDefaultIocp) {
            this.iocp.detachFromThreadPool();
        }
    }
    
    @Override
    public AsynchronousChannelGroupImpl group() {
        return this.iocp;
    }
    
    private static IOException toIOException(Throwable cause) {
        if ((cause instanceof IOException)) {
            if ((cause instanceof ClosedChannelException)) {
                cause = new AsynchronousCloseException();
            }
            return (IOException)cause;
        }
        else {
            return new IOException(cause);
        }
    }
    
    @Override
    public long size() throws IOException {
        try {
            this.begin();
            return WindowsAsynchronousFileChannelImpl.nd.size(this.fdObj);
        }
        finally {
            this.end();
        }
    }
    
    @Override
    public AsynchronousFileChannel truncate(long n) throws IOException {
        if (n < 0L) {
            throw new IllegalArgumentException("Negative size");
        }
        if (!this.writing) {
            throw new NonWritableChannelException();
        }
        try {
            this.begin();
            if (n > WindowsAsynchronousFileChannelImpl.nd.size(this.fdObj)) {
                return this;
            }
            WindowsAsynchronousFileChannelImpl.nd.truncate(this.fdObj, n);
        }
        finally {
            this.end();
        }
        return this;
    }
    
    @Override
    public void force(boolean b) throws IOException {
        try {
            this.begin();
            WindowsAsynchronousFileChannelImpl.nd.force(this.fdObj, b);
        }
        finally {
            this.end();
        }
    }
    
    @Override
     <A> Future<FileLock> implLock(long n, long n2, boolean b, A a, CompletionHandler<FileLock, ? super A> completionHandler) {
        if (b && !this.reading) {
            throw new NonReadableChannelException();
        }
        if (!b && !this.writing) {
            throw new NonWritableChannelException();
        }
        FileLockImpl addToFileLockTable = this.addToFileLockTable(n, n2, b);
        if (addToFileLockTable == null) {
            ClosedChannelException ex = new ClosedChannelException();
            if (completionHandler == null) {
                return CompletedFuture.withFailure(ex);
            }
            Invoker.invoke(this, completionHandler, a, null, ex);
            return null;
        }
        else {
            PendingFuture pendingFuture = new PendingFuture(this, completionHandler, a);
            WindowsAsynchronousFileChannelImpl.LockTask<Object> context = new WindowsAsynchronousFileChannelImpl.LockTask(n, addToFileLockTable, pendingFuture);
            pendingFuture.setContext(context);
            if (Iocp.supportsThreadAgnosticIo()) {
                context.run();
            }
            else {
                boolean n3 = false;
                try {
                    Invoker.invokeOnThreadInThreadPool(this, context);
                    n3 = true;
                }
                finally {
                    if (n3 == 0) {
                        this.removeFromFileLockTable(addToFileLockTable);
                    }
                }
            }
            return pendingFuture;
        }
    }
    
    @Override
    public FileLock tryLock(long n, long n2, boolean b) throws IOException {
        if (b && !this.reading) {
            throw new NonReadableChannelException();
        }
        if (!b && !this.writing) {
            throw new NonWritableChannelException();
        }
        FileLockImpl addToFileLockTable = this.addToFileLockTable(n, n2, b);
        if (addToFileLockTable == null) {
            throw new ClosedChannelException();
        }
        int n3 = 0;
        try {
            this.begin();
            if (WindowsAsynchronousFileChannelImpl.nd.lock(this.fdObj, false, n, n2, b) == -1) {
                return null;
            }
            n3 = 1;
            return addToFileLockTable;
        }
        finally {
            if (n3 == 0) {
                this.removeFromFileLockTable(addToFileLockTable);
            }
            this.end();
        }
    }
    
    @Override
    protected void implRelease(FileLockImpl fileLockImpl) throws IOException {
        WindowsAsynchronousFileChannelImpl.nd.release(this.fdObj, fileLockImpl.position(), fileLockImpl.size());
    }
    
    @Override
     <A> Future<Integer> implRead(ByteBuffer byteBuffer, long n, A a, CompletionHandler<Integer, ? super A> completionHandler) {
        if (!this.reading) {
            throw new NonReadableChannelException();
        }
        if (n < 0L) {
            throw new IllegalArgumentException("Negative position");
        }
        if (byteBuffer.isReadOnly()) {
            throw new IllegalArgumentException("Read-only buffer");
        }
        if (!this.isOpen()) {
            ClosedChannelException ex = new ClosedChannelException();
            if (completionHandler == null) {
                return CompletedFuture.withFailure(ex);
            }
            Invoker.invoke(this, completionHandler, a, null, ex);
            return null;
        }
        else {
            int position = byteBuffer.position();
            int limit = byteBuffer.limit();
            assert position <= limit;
            int n2 = (position <= limit) ? (limit - position) : 0;
            if (n2 == 0) {
                if (completionHandler == null) {
                    return CompletedFuture.withResult(Integer.valueOf(0));
                }
                Invoker.invoke(this, completionHandler, a, Integer.valueOf(0), null);
                return null;
            }
            else {
                PendingFuture pendingFuture = new PendingFuture(this, completionHandler, a);
                WindowsAsynchronousFileChannelImpl.ReadTask<Object> context = new WindowsAsynchronousFileChannelImpl.ReadTask(byteBuffer, position, n2, n, pendingFuture);
                pendingFuture.setContext(context);
                if (Iocp.supportsThreadAgnosticIo()) {
                    context.run();
                    return pendingFuture;
                }
                Invoker.invokeOnThreadInThreadPool(this, context);
                return pendingFuture;
            }
        }
    }
    
    @Override
     <A> Future<Integer> implWrite(ByteBuffer byteBuffer, long n, A a, CompletionHandler<Integer, ? super A> completionHandler) {
        if (!this.writing) {
            throw new NonWritableChannelException();
        }
        if (n < 0L) {
            throw new IllegalArgumentException("Negative position");
        }
        if (!this.isOpen()) {
            ClosedChannelException ex = new ClosedChannelException();
            if (completionHandler == null) {
                return CompletedFuture.withFailure(ex);
            }
            Invoker.invoke(this, completionHandler, a, null, ex);
            return null;
        }
        else {
            int position = byteBuffer.position();
            int limit = byteBuffer.limit();
            assert position <= limit;
            int n2 = (position <= limit) ? (limit - position) : 0;
            if (n2 == 0) {
                if (completionHandler == null) {
                    return CompletedFuture.withResult(Integer.valueOf(0));
                }
                Invoker.invoke(this, completionHandler, a, Integer.valueOf(0), null);
                return null;
            }
            else {
                PendingFuture pendingFuture = new PendingFuture(this, completionHandler, a);
                WindowsAsynchronousFileChannelImpl.WriteTask<Object> context = new WindowsAsynchronousFileChannelImpl.WriteTask(byteBuffer, position, n2, n, pendingFuture);
                pendingFuture.setContext(context);
                if (Iocp.supportsThreadAgnosticIo()) {
                    context.run();
                    return pendingFuture;
                }
                Invoker.invokeOnThreadInThreadPool(this, context);
                return pendingFuture;
            }
        }
    }
    
    private static native int readFile(long p0, long p1, int p2, long p3, long p4) throws IOException;
    
    private static native int writeFile(long p0, long p1, int p2, long p3, long p4) throws IOException;
    
    private static native int lockFile(long p0, long p1, long p2, boolean p3, long p4) throws IOException;
    
    private static native void close0(long p0);
    
    static int access$200(long n, long n2, long n3, boolean b, long n4) throws IOException {
        return lockFile(n, n2, n3, b, n4);
    }
    
    static IOException access$300(Throwable t) {
        return toIOException(t);
    }
    
    static int access$400(long n, long n2, int n3, long n4, long n5) throws IOException {
        return readFile(n, n2, n3, n4, n5);
    }
    
    static int access$500(long n, long n2, int n3, long n4, long n5) throws IOException {
        return writeFile(n, n2, n3, n4, n5);
    }
    
    static {
        fdAccess = SharedSecrets.getJavaIOFileDescriptorAccess();
        nd = new FileDispatcherImpl();
        IOUtil.load();
    }
    
    private static class DefaultIocpHolder
    {
        static final Iocp defaultIocp;
        
        private static Iocp defaultIocp() {
            try {
                return new Iocp(null, ThreadPool.createDefault()).start();
            }
            catch (IOException cause) {
                throw new InternalError(cause);
            }
        }
        
        static {
            defaultIocp = defaultIocp();
        }
    }
    
    private class LockTask<A> implements Runnable, Iocp.ResultHandler
    {
        private final long position;
        private final FileLockImpl fli;
        private final PendingFuture<FileLock, A> result;
        
        LockTask(long position, FileLockImpl fli, PendingFuture<FileLock, A> result) {
            this.position = position;
            this.fli = fli;
            this.result = result;
        }
        
        @Override
        public void run() {
            long add = 0L;
            boolean n = false;
            try {
                WindowsAsynchronousFileChannelImpl.this.begin();
                add = WindowsAsynchronousFileChannelImpl.this.ioCache.add(this.result);
                synchronized (this.result) {
                    if (WindowsAsynchronousFileChannelImpl.access$200(WindowsAsynchronousFileChannelImpl.this.handle, this.position, this.fli.size(), this.fli.isShared(), add) == -2) {
                        n = true;
                        return;
                    }
                    this.result.setResult(this.fli);
                }
            }
            catch (Throwable t) {
                WindowsAsynchronousFileChannelImpl.this.removeFromFileLockTable(this.fli);
                this.result.setFailure(WindowsAsynchronousFileChannelImpl.access$300(t));
            }
            finally {
                if (n == 0 && add != 0L) {
                    WindowsAsynchronousFileChannelImpl.this.ioCache.remove(add);
                }
                WindowsAsynchronousFileChannelImpl.this.end();
            }
            Invoker.invoke(this.result);
        }
        
        @Override
        public void completed(int n, boolean b) {
            this.result.setResult(this.fli);
            if (b) {
                Invoker.invokeUnchecked(this.result);
            }
            else {
                Invoker.invoke(this.result);
            }
        }
        
        @Override
        public void failed(int n, IOException failure) {
            WindowsAsynchronousFileChannelImpl.this.removeFromFileLockTable(this.fli);
            if (WindowsAsynchronousFileChannelImpl.this.isOpen()) {
                this.result.setFailure(failure);
            }
            else {
                this.result.setFailure(new AsynchronousCloseException());
            }
            Invoker.invoke(this.result);
        }
    }
    
    private class ReadTask<A> implements Runnable, Iocp.ResultHandler
    {
        private final ByteBuffer dst;
        private final int pos;
        private final int rem;
        private final long position;
        private final PendingFuture<Integer, A> result;
        private volatile ByteBuffer buf;
        
        ReadTask(ByteBuffer dst, int pos, int rem, long position, PendingFuture<Integer, A> result) {
            this.dst = dst;
            this.pos = pos;
            this.rem = rem;
            this.position = position;
            this.result = result;
        }
        
        void releaseBufferIfSubstituted() {
            if (this.buf != this.dst) {
                Util.releaseTemporaryDirectBuffer(this.buf);
            }
        }
        
        void updatePosition(int n) {
            if (n > 0) {
                if (this.buf == this.dst) {
                    try {
                        this.dst.position(this.pos + n);
                    }
                    catch (IllegalArgumentException ex2) {}
                }
                else {
                    this.buf.position(n).flip();
                    try {
                        this.dst.put(this.buf);
                    }
                    catch (BufferOverflowException ex) {}
                }
            }
        }
        
        @Override
        public void run() {
            long add = 0L;
            long address;
            if ((this.dst instanceof DirectBuffer)) {
                this.buf = this.dst;
                address = ((DirectBuffer)this.dst).address() + (long)this.pos;
            }
            else {
                this.buf = Util.getTemporaryDirectBuffer(this.rem);
                address = ((DirectBuffer)this.buf).address();
            }
            boolean n = false;
            try {
                WindowsAsynchronousFileChannelImpl.this.begin();
                add = WindowsAsynchronousFileChannelImpl.this.ioCache.add(this.result);
                int access$400 = WindowsAsynchronousFileChannelImpl.access$400(WindowsAsynchronousFileChannelImpl.this.handle, address, this.rem, this.position, add);
                if (access$400 == -2) {
                    n = true;
                    return;
                }
                if (access$400 == -1) {
                    this.result.setResult(Integer.valueOf(access$400));
                }
                else {
                    throw new InternalError("Unexpected result: " + access$400);
                }
            }
            catch (Throwable t) {
                this.result.setFailure(WindowsAsynchronousFileChannelImpl.access$300(t));
            }
            finally {
                if (n == 0) {
                    if (add != 0L) {
                        WindowsAsynchronousFileChannelImpl.this.ioCache.remove(add);
                    }
                    this.releaseBufferIfSubstituted();
                }
                WindowsAsynchronousFileChannelImpl.this.end();
            }
            Invoker.invoke(this.result);
        }
        
        @Override
        public void completed(int i, boolean b) {
            this.updatePosition(i);
            this.releaseBufferIfSubstituted();
            this.result.setResult(Integer.valueOf(i));
            if (b) {
                Invoker.invokeUnchecked(this.result);
            }
            else {
                Invoker.invoke(this.result);
            }
        }
        
        @Override
        public void failed(int n, IOException failure) {
            if (n == 38) {
                this.completed(-1, false);
            }
            else {
                this.releaseBufferIfSubstituted();
                if (WindowsAsynchronousFileChannelImpl.this.isOpen()) {
                    this.result.setFailure(failure);
                }
                else {
                    this.result.setFailure(new AsynchronousCloseException());
                }
                Invoker.invoke(this.result);
            }
        }
    }
    
    private class WriteTask<A> implements Runnable, Iocp.ResultHandler
    {
        private final ByteBuffer src;
        private final int pos;
        private final int rem;
        private final long position;
        private final PendingFuture<Integer, A> result;
        private volatile ByteBuffer buf;
        
        WriteTask(ByteBuffer src, int pos, int rem, long position, PendingFuture<Integer, A> result) {
            this.src = src;
            this.pos = pos;
            this.rem = rem;
            this.position = position;
            this.result = result;
        }
        
        void releaseBufferIfSubstituted() {
            if (this.buf != this.src) {
                Util.releaseTemporaryDirectBuffer(this.buf);
            }
        }
        
        void updatePosition(int n) {
            if (n > 0) {
                try {
                    this.src.position(this.pos + n);
                }
                catch (IllegalArgumentException ex) {}
            }
        }
        
        @Override
        public void run() {
            long add = 0L;
            long address;
            if ((this.src instanceof DirectBuffer)) {
                this.buf = this.src;
                address = ((DirectBuffer)this.src).address() + (long)this.pos;
            }
            else {
                (this.buf = Util.getTemporaryDirectBuffer(this.rem)).put(this.src);
                this.buf.flip();
                this.src.position(this.pos);
                address = ((DirectBuffer)this.buf).address();
            }
            try {
                WindowsAsynchronousFileChannelImpl.this.begin();
                add = WindowsAsynchronousFileChannelImpl.this.ioCache.add(this.result);
                int access$500 = WindowsAsynchronousFileChannelImpl.access$500(WindowsAsynchronousFileChannelImpl.this.handle, address, this.rem, this.position, add);
                if (access$500 == -2) {
                    return;
                }
                throw new InternalError("Unexpected result: " + access$500);
            }
            catch (Throwable t) {
                this.result.setFailure(WindowsAsynchronousFileChannelImpl.access$300(t));
                if (add != 0L) {
                    WindowsAsynchronousFileChannelImpl.this.ioCache.remove(add);
                }
                this.releaseBufferIfSubstituted();
            }
            finally {
                WindowsAsynchronousFileChannelImpl.this.end();
            }
            Invoker.invoke(this.result);
        }
        
        @Override
        public void completed(int i, boolean b) {
            this.updatePosition(i);
            this.releaseBufferIfSubstituted();
            this.result.setResult(Integer.valueOf(i));
            if (b) {
                Invoker.invokeUnchecked(this.result);
            }
            else {
                Invoker.invoke(this.result);
            }
        }
        
        @Override
        public void failed(int n, IOException failure) {
            this.releaseBufferIfSubstituted();
            if (WindowsAsynchronousFileChannelImpl.this.isOpen()) {
                this.result.setFailure(failure);
            }
            else {
                this.result.setFailure(new AsynchronousCloseException());
            }
            Invoker.invoke(this.result);
        }
    }
}
